using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class WorldGenerator : MonoBehaviour
{
	public enum eGenerateMode
	{
		MODE_NOTHING,
		MODE_ROOMS,
		MODE_CONNECT,
		MODE_FILL_WALLS,
		MODE_ADD_ACTIONS,
		MODE_ADD_MEANING
	};

	eGenerateMode m_mode = eGenerateMode.MODE_NOTHING;

	public GameObject m_wall;
	public GameObject m_floor;
	public GameObject m_actionPrefab;

	Vector2 m_startPos = new Vector2(0, 0);
	Vector2 m_mapSize = new Vector2(100, 100);
	public List<Room> m_rooms = new List<Room>();
	public List<Cell> m_cells = new List<Cell>();

	public List<Sprite> m_floors = new List<Sprite>();
	GameObject m_level;
	int m_actionsToAdd = -1;
    int m_fillWallsCur = 0;
	int m_roomsLeft = 0;


	const float cameraZoomMapGen = 60;
	int m_curRoomID = 0;

	Cell GetCell(int x, int y)
	{
		return m_cells[x + y * (int)m_mapSize.x];
	}

	void CopyToCell(Cell source, Cell dest)
	{
		source.m_type = dest.m_type;
	}

	void SetCellHorizontal(int posX, int posY, ref Cell sourceCell, int size)
	{
		for (int i= 0; i < size; i++)
		{
			Cell pCell = GetCell(posX + i, posY);
			CopyToCell(pCell, sourceCell);
		}
	}

	void SetCellVertical(int posX, int posY, ref Cell sourceCell, int size)
	{
		for (int i = 0; i < size; i++)
		{
			Cell pCell = GetCell(posX, posY+i);
			CopyToCell(pCell, sourceCell);
        }
	}

	void UpdateVisualWithCell(int cellX, int cellY)
	{

		Cell pCell = GetCell(cellX, cellY);

		Vector3 vPos = new Vector3(cellX, m_mapSize.y-cellY, 0);

		GameObject obj = m_floor;

		switch(pCell.m_type)
		{
			case Cell.eCellType.TYPE_WALL:
				obj = m_wall;
				break;

			case Cell.eCellType.TYPE_HALLWAY:
				obj = m_floor;
				break;
		}

		if (pCell.m_type != Cell.eCellType.TYPE_EMPTY)
		{
			GameObject o = (GameObject)Instantiate(obj, vPos, Quaternion.identity);
			o.transform.parent = m_level.transform;

			SpriteRenderer r = o.GetComponentInChildren<SpriteRenderer>();
			if (pCell.m_type == Cell.eCellType.TYPE_FLOOR)
			{
				r.sprite = m_floors[ m_rooms[pCell.m_parentRoomID].m_floorTextureIndex];
			}
		}

	}

	public int MapToWorldY(int y)
	{
		return (int)m_mapSize.y - y;
	}

	public void GenerateWorld()
	{
		m_level = new GameObject("Level");
		GameLogic.Get().GetPlayer().transform.position = new Vector3(m_mapSize.x / 2, (float)MapToWorldY((int)m_mapSize.y/2), 0);
		GameLogic.Get().GetCamera().orthographicSize = cameraZoomMapGen;
		//add border
		m_rooms.Add(new Room());
		m_rooms[0].m_name = "Outside";

		for (int i = 0; i < m_mapSize.x * m_mapSize.y; i++)
		{
			m_cells.Add(new Cell()); //gotta be faster way to do this?
		}

		Cell wallCell = new Cell();
		wallCell.m_type = Cell.eCellType.TYPE_WALL;
		//add border
		SetCellHorizontal(0, 0, ref wallCell, (int)m_mapSize.x);
		SetCellHorizontal(0, (int)m_mapSize.y-1, ref wallCell, (int)m_mapSize.x);

		SetCellVertical(0, 0, ref wallCell, (int)m_mapSize.y);
		SetCellVertical((int)m_mapSize.y-1, 0, ref wallCell, (int)m_mapSize.y);

		m_roomsLeft = (int)(m_mapSize.x * m_mapSize.y) / 100;
        Debug.Log("Cell count: " + m_cells.Count);
		m_cells[0].m_type = Cell.eCellType.TYPE_WALL;
	
		//show our data a bit

		for (int y = 0; y < (int)m_mapSize.y; y++)
		{
			string s = "";
			for (int x = 0; x < (int)m_mapSize.x; x++)
			{
				//s = s + " "+(int)m_cells[x+y*(int)m_mapSize.x].m_type;
				UpdateVisualWithCell(x, y);
			}
			//Debug.Log(s);
		}
		m_mode = eGenerateMode.MODE_ROOMS;
	}

	bool CellIsOpenForRoom(int roomID, int x, int y)
	{
		Cell pCell = GetCell(x, y);
		if (pCell.m_parentRoomID != -1) return false;
		if (x < 2 || x-2 >= (int)m_mapSize.x
			|| y < 2 || y-2 >= (int) m_mapSize.y)
		{
			return false; //touching an edge
		}
		return true;
	}

	bool CreateRoom(int roomID, int posX, int posY)
	{
	
		int roomSizeX = Random.Range(3, 8);
		int roomSizeY = Random.Range(3, 8);

		posX -= roomSizeX / 2;
		posY -= roomSizeY / 2;

		bool bFailed = false;

		for (int x = 0; x < roomSizeX; x++)
		{
			for (int y= 0; y < roomSizeY; y++)
			{
				if (!CellIsOpenForRoom(roomID, posX+x, posY+y))
				{
					bFailed = true;
				}

			}
		}

		if (bFailed)
		{
			return false;
		}

		//add it!

		m_rooms.Add(new Room());
		m_rooms[roomID].m_name = "Room " + roomID;
		m_rooms[roomID].m_centerX = posX;
		m_rooms[roomID].m_centerY = posY;
		m_rooms[roomID].m_floorTextureIndex = Random.Range((int)1, (int)5);

		for (int x = 0; x < roomSizeX; x++)
		{
			for (int y = 0; y < roomSizeY; y++)
			{
				Cell pCell = GetCell(posX + x, posY + y);
				pCell.m_type = Cell.eCellType.TYPE_FLOOR;
				pCell.m_parentRoomID = roomID;
				UpdateVisualWithCell(posX + x, posY + y);
			}
		}
		

		return true;

	}

	void ModeAddRooms()
	{
		if (m_roomsLeft > 0)
		{
			m_roomsLeft--;
			m_curRoomID++; //0 is reserved for no room, so we start at 1
			//Debug.Log("Generating room " + m_curRoomID);
			if (m_curRoomID == 1)
			{
				//special case for first room
				CreateRoom(m_curRoomID, (int)m_mapSize.x / 2, (int)m_mapSize.y / 2);
			}
			else
			{
				int tries = 30;
				while (tries-- > 0)
				{

					if (CreateRoom(m_curRoomID, Random.Range(6, (int)m_mapSize.x - 6), Random.Range(6, (int)m_mapSize.y - 6)))
					{
						//success
						break;
					}

				}
			}


		}
		else
		{
			if (m_roomsLeft != -1)
			{
				//Debug.Log("Finished adding rooms");
				m_mode = eGenerateMode.MODE_CONNECT;
				m_roomsLeft = m_rooms.Count-1;
				m_curRoomID = 0;
				Debug.Log("Room count: " + m_rooms.Count);

			}

		}
	}

	void MoveTowardsTarget(ref int num, int target, int amount)
	{
		if (num < target) num += amount; else num -= amount;
	}

	void ConnectRooms(int sourceRoom, int destRoom)
	{
		//Debug.Log("Connecting room " + sourceRoom + " and room " + destRoom);
		int curX = m_rooms[sourceRoom].m_centerX;
		int curY = m_rooms[sourceRoom].m_centerY;

		int destX = m_rooms[destRoom].m_centerX;
		int destY = m_rooms[destRoom].m_centerY;
        while (curX != destX
			|| curY != destY)
		{
			//we aren't there yet, so dig a tunnel if we have to
			bool bUseX = Random.Range((int)0, (int)2) == 0;

			//Debug.Log("Chose " + bUseX);

			if (bUseX && curX == destX)
			{
				//we're already there, use the other axis
				bUseX = !bUseX;
			}
			if (!bUseX && curY == destY)
			{
				//we're already there, use the other axis
				bUseX = !bUseX;
			}

			if (bUseX) MoveTowardsTarget(ref curX, destX, 1);
			if (!bUseX) MoveTowardsTarget(ref curY, destY, 1);

			Cell pCell = GetCell(curX, curY);

			if (pCell.m_parentRoomID == -1)
			{
				//it's unowned!  Knock out this wall
				pCell.m_type = Cell.eCellType.TYPE_HALLWAY;
				pCell.m_parentRoomID = 0;
				UpdateVisualWithCell(curX, curY);
			} else
			{
				if (pCell.m_parentRoomID != sourceRoom)
				{
					//abort early, we connected with... something
					return;
				}
			}
			
		}

	}

	bool CellIsTouchingRoomOrHallway(int x, int y)
	{
		//ignore the ones by the borders so we don't get bad accesses
		if ( x < 1 || y < 1) return false;
		if (x+2 >= m_mapSize.x || y+2 >= m_mapSize.y) return false;

		Cell pCell = GetCell(x - 1, y);
		if (pCell.m_parentRoomID != -1) return true;

		pCell = GetCell(x + 1, y);
		if (pCell.m_parentRoomID != -1) return true;

		pCell = GetCell(x, y-1);
		if (pCell.m_parentRoomID != -1) return true;

		pCell = GetCell(x, y+1);
		if (pCell.m_parentRoomID != -1) return true;

		return false;
	}

	void FillWalls()
	{

		if (m_fillWallsCur >= m_cells.Count)
		{
			//all done
			m_mode = eGenerateMode.MODE_ADD_ACTIONS;
			m_actionsToAdd = -1;
			return;
		}

		int count = 0;
		while (m_fillWallsCur < m_cells.Count && count < 300)
		{
		
			if (m_cells[m_fillWallsCur].m_parentRoomID == -1)
			{
				//unknowned
				int y = m_fillWallsCur / (int)m_mapSize.x;
				int x = m_fillWallsCur - (y * (int)m_mapSize.x);

				//Debug.Log("Setting " + x + " " + y);

				//does it border a room/hallway?
				if (CellIsTouchingRoomOrHallway(x, y))
				{
					//set to wall
					//m_cells[m_fillWallsCur].m_parentRoomID = 0;
					m_cells[m_fillWallsCur].m_type = Cell.eCellType.TYPE_WALL;
					UpdateVisualWithCell(x, y);
				}

			}

			m_fillWallsCur++;
			count++;
		}

	}

	void ModeAddConnections()
	{
		m_curRoomID++; //will start at 1
		
		if (m_curRoomID >= m_rooms.Count)
		{
			//all done
			m_fillWallsCur = 0;
			m_mode = eGenerateMode.MODE_FILL_WALLS;
		} else
		{
			int tries = 2;

			while (tries-- > 0)
			{

				int sourceRoom = m_curRoomID;
				int destRoom = m_rooms.Count - m_curRoomID;
				if (destRoom == sourceRoom || destRoom < 1 || destRoom >= m_rooms.Count) destRoom = Random.Range(1, m_rooms.Count);
				if (destRoom == sourceRoom) destRoom = 3; //weird but whatever, this should be very rare

				ConnectRooms(sourceRoom, destRoom);
			}
		} 
	}

	void AddActions()
	{
		if (m_actionsToAdd == 0)
		{ 
			//done
			m_mode = eGenerateMode.MODE_ADD_MEANING;
			return;
		}

		if (m_actionsToAdd == -1)
		{
			//first time here
			m_actionsToAdd = m_rooms.Count * 2;
			//m_actionsToAdd = 10;
		}

		int x = Random.Range((int)3, (int)m_mapSize.x - 3);
		int y = Random.Range((int)3, (int)m_mapSize.y - 3);

		Vector2 vCenter = new Vector2(m_mapSize.x / 2, m_mapSize.y/2);
		Vector2 vPoint = new Vector2(x, y);

		float dist= (vCenter - vPoint).magnitude;
		//GameLogic.Log("That's " + dist + " from center");
		float diff = dist / (m_mapSize.x/2); //0 means close to middle, 1 means close to edge of map

		Cell pCell = GetCell(x, y);
		if (pCell.m_itemPlacedHere) return;

		Cell pCellBelow = GetCell(x, y+1);
		if (pCellBelow.m_itemPlacedHere) return; //looks bad because the text crowds eachother
		pCellBelow = GetCell(x, y - 1);
		if (pCellBelow.m_itemPlacedHere) return; //looks bad because the text crowds eachother

		if (pCell.m_type == Cell.eCellType.TYPE_FLOOR)
		{
			pCell.m_itemPlacedHere = true;
			m_actionsToAdd--;
			//Vector2 v(x-m_, y);
			//Debug.Log("Creating item!");
			//float diff = 
			GameObject o = (GameObject)Instantiate(m_actionPrefab, new Vector3(x,m_mapSize.y-y,0), Quaternion.identity);
			o.transform.parent = m_level.transform;
			Action actionScript = o.GetComponent<Action>();

			int r = Random.Range((int)0, (int)2);
			switch(r)
			{
				case 0:
					actionScript.SetupRandomJob(diff);
					break;
				case 1:
					actionScript.SetupRandomActivity(diff);
					break;
			}
			return;
		}

		if (pCell.m_type == Cell.eCellType.TYPE_HALLWAY)
		{

			pCell.m_itemPlacedHere = true;
			m_actionsToAdd--;

			GameObject o = (GameObject)Instantiate(m_actionPrefab, new Vector3(x, m_mapSize.y - y, 0), Quaternion.identity);
			o.transform.parent = m_level.transform;
			Action actionScript = o.GetComponent<Action>();

			int r = Random.Range((int)0, (int)1);

			switch (r)
			{
				case 0:
					actionScript.SetupRandomGateway(diff);
					break;
			}

			return;
		}

	
	}

	void AddMeaning(int meaningNum)
	{
		//done

		while(true)
		{

			int x = Random.Range((int)3, (int)m_mapSize.x - 3);
			int y = Random.Range((int)3, (int)m_mapSize.y - 3);

			Vector2 vCenter = new Vector2(m_mapSize.x / 2, m_mapSize.y / 2);
			Vector2 vPoint = new Vector2(x, y);

			float dist = (vCenter - vPoint).magnitude;
			//GameLogic.Log("That's " + dist + " from center");
			float diff = dist / (m_mapSize.x / 2); //0 means close to middle, 1 means close to edge of map

			if (diff < 0.6) continue; //try again, too close to middle
			//if (diff > 0.1) continue; //try again, too close to middle

			Cell pCell = GetCell(x, y);
			if (pCell.m_itemPlacedHere) continue; //try again, too close to something else

			Cell pCellBelow = GetCell(x, y + 1);
			if (pCellBelow.m_itemPlacedHere) continue; //try again, too close to something else
			pCellBelow = GetCell(x, y - 1);
			if (pCellBelow.m_itemPlacedHere) continue; //try again, too close to something else

			if (pCell.m_type != Cell.eCellType.TYPE_FLOOR) continue; //has to be a room

			pCell.m_itemPlacedHere = true;

			//ok, this looks like a good place
			GameObject o = (GameObject)Instantiate(m_actionPrefab, new Vector3(x, m_mapSize.y - y, 0), Quaternion.identity);
			o.transform.parent = m_level.transform;
			Action actionScript = o.GetComponent<Action>();
			actionScript.SetupMeaning(meaningNum);
			break; //the only we get out of here.  Yeah, never program like this
		}
	//	tryAgain;




	}

	void FixedUpdate()
	{

		switch(m_mode)
		{

			case eGenerateMode.MODE_ROOMS:
				ModeAddRooms();
				if (m_mode == eGenerateMode.MODE_ROOMS)
					ModeAddRooms();
				if (m_mode == eGenerateMode.MODE_ROOMS)
					ModeAddRooms();
				break;

			case eGenerateMode.MODE_CONNECT:
				ModeAddConnections();
				if (m_mode == eGenerateMode.MODE_CONNECT)
					ModeAddConnections();
				if (m_mode == eGenerateMode.MODE_CONNECT)
					ModeAddConnections();

				break;

			case eGenerateMode.MODE_FILL_WALLS:
				FillWalls();
				break;

			case eGenerateMode.MODE_ADD_ACTIONS:

				int actionsPerCycle = 40;
				for (int i=0; i < actionsPerCycle; i++)
				{
					if (m_mode != eGenerateMode.MODE_ADD_ACTIONS) break;
					AddActions();
				}
				break;

			case eGenerateMode.MODE_ADD_MEANING:

				AddMeaning(0); //one shot, don't want the player to see this really
				AddMeaning(1); //one shot, don't want the player to see this really
				AddMeaning(2); //one shot, don't want the player to see this really
				AddMeaning(3); //one shot, don't want the player to see this really
				AddMeaning(4); //one shot, don't want the player to see this really
				m_mode = eGenerateMode.MODE_NOTHING;
				GameLogic.Get().StartGame();

				break;

		}
		

	}

}

